from pathlib import Path
from typing import Tuple

# Get git root
git_root = Path(__file__).parent.parent

no_git_str = '.git structure does not exist'


class GitInfo:
    """
    Provides ``git`` repo information

    :param input_path: Full path containing the ``.git`` contents

    :ivar input_path: Full path containing the ``.git`` contents
    :ivar head_path: Full path of the ``.git`` HEAD
    :ivar branch: Active branch name
    :ivar commit: Full hash
    :ivar short_commit: short hash commit
    """
    def __init__(self, input_path: str):
        self.input_path: str = input_path
        self.head_path: Path = Path(self.input_path) / ".git" / "HEAD"
        self.branch: str = self.get_active_branch_name()
        commit_tuple: tuple = self.get_latest_commit()
        self.commit: str = commit_tuple[0]
        self.short_commit: str = commit_tuple[1]

    def get_active_branch_name(self) -> str:
        """Retrieve active branch name"""
        if self.head_path.exists():
            with self.head_path.open("r") as f:
                content = f.read().splitlines()

            for line in content:
                if line[0:4] == "ref:":
                    return line.partition("refs/heads/")[2]
                else:
                    return f"HEAD detached : {content[0]}"
        else:
            return no_git_str

    def get_latest_commit(self) -> Tuple[str, str]:
        """Retrieve latest commit hash"""
        if self.head_path.exists():
            with self.head_path.open("r") as f:
                content = f.read().splitlines()

            for line in content:
                if line[0:4] == "ref:":
                    ref_path = Path(self.input_path) / ".git" / f"{line.partition(' ')[2]}"
                    with ref_path.open('r') as g:
                        commit = g.read().splitlines()
                else:
                    commit = content

            return commit[0], commit[0][:7]  # full and short hash
        else:
            return no_git_str, ''



import json
from typing import Union

import requests

from .logger import log_stdout
from logging import Logger

method_list = ['GET', 'PUT', 'POST', 'DELETE']


def redata_request(method: str, url: str, headers: dict, data: dict = None,
                   binary: bool = False, params: dict = None,
                   process: bool = True, log: Logger = log_stdout()) \
        -> Union[dict, requests.Response]:
    """
    Wrapper for common HTTP requests

    This code is based of a function in cognoma's figshare:
    https://github.com/cognoma/figshare

    :param method: HTTP method. One of GET, PUT, POST or DELETE
    :param url: URL for the request
    :param headers: HTTP header information
    :param data: HTTP data for PUT/POST
    :param binary: Whether data is binary or not
    :param params: Additional information for URL GET request
    :param process: Returns JSON content, otherwise the full request
                    is provided. Default: True
    :param log: Logger object for stdout and file logging. Default: stdout

    :return: JSON response for the request returned as python dict or
             the full request. See ``process`` option
    """

    if method not in method_list:
        log.warning("Incorrect method input provided!")
        log.warning(f"Must be one of these: {', '.join(method_list)}")
        raise ValueError

    if data is not None and not binary:
        data = json.dumps(data)

    response = requests.request(method, url, headers=headers,
                                data=data, params=params)

    if process:
        try:
            response.raise_for_status()
            try:
                response_data = json.loads(response.text)
            except ValueError:
                response_data = response.content
        except requests.exceptions.HTTPError as error:
            log.warning(f'Caught an HTTPError: {error}')
            log.warning('Body:\n', response.text)
            raise

        return response_data
    else:
        return response



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import sphinx_rtd_theme
import sphinx_autodoc_typehints
sys.path.insert(0, os.path.abspath('../../'))

# -- Project information -----------------------------------------------------

project = 'redata-commons'
copyright = '2021, Arizona Board of Regents'
author = 'Chun Ly, UA Research Data Repository (ReDATA) Team'

# The full version, including alpha/beta/rc tags
release = 'v0.2.0'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_rtd_theme",
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinx.ext.extlinks",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# -- External links for repeated use:
repo_link = 'https://github.com/UAL-RE/redata-commons'
extlinks = {
    'repo': (f"{repo_link}/%s", None),
    'repo-main-file': (f"{repo_link}/blob/main/%s", None),
    'pypi': ('https://pypi.org/project/redata/%s', None),
}



import sys
import io
from os import path, uname, chmod, mkdir

from datetime import date
import pandas as pd

import logging

# User and hostname
from getpass import getuser
from socket import gethostname

from requests import get

from redata import __version__
from redata.commons.git_info import GitInfo

today = date.today()

formatter = logging.Formatter('%(asctime)s - %(levelname)8s: %(message)s', "%H:%M:%S")

file_formatter = logging.Formatter('%(asctime)s %(levelname)8s - %(module)20s %(funcName)30s : %(message)s',
                                   "%H:%M:%S")


class LogClass:
    """
    Main class to log information to stdout and ASCII logfile

    :param log_dir: Relative path for exported logfile directory
    :param logfile: Filename for exported log file

    :ivar LOG_FILENAME: Full path of log file
    :ivar file_log_level: File log level: DEBUG

    To use: ``log = LogClass(log_dir, logfile).get_logger()``
    """

    def __init__(self, log_dir: str, logfile: str):
        self.LOG_FILENAME: str = path.join(log_dir, logfile)
        self.file_log_level = logging.DEBUG  # For file logging

    def get_logger(self):
        """Primary method to retrieve stdout and ASCII file Logging object"""
        log = logging.getLogger("main_logger")
        if not log.handlers:
            log.setLevel(self.file_log_level)

            sh = logging.StreamHandler(sys.stdout)
            sh.setLevel(logging.INFO)  # Only at INFO level
            sh.setFormatter(formatter)
            log.addHandler(sh)

            fh = logging.FileHandler(self.LOG_FILENAME)
            fh.setLevel(self.file_log_level)
            fh.setFormatter(file_formatter)
            log.addHandler(fh)

            log.handler_set = True
            log.propagate = False
        return log


def log_stdout() -> logging.Logger:
    """
    Stdout logger

    :return: ``log``
    """
    log_level = logging.INFO
    log = logging.getLogger("stdout_logger")
    if not log.handlers:
        log.setLevel(log_level)
        sh = logging.StreamHandler(sys.stdout)
        sh.setFormatter(formatter)
        log.addHandler(sh)

        log.handler_set = True
        log.propagate = False
    return log


def log_setup(log_dir: str, logfile_prefix: str) -> logging.Logger:
    """
    Create Logger object (``log``) for stdout and file logging

    :param log_dir: Directory for logs
    :param logfile_prefix: Log file prefix

    :return: Logger object
    """
    if not path.exists(log_dir):
        mkdir(log_dir)
    logfile = f'{logfile_prefix}.{today.strftime("%Y-%m-%d")}.log'

    log = LogClass(log_dir, logfile).get_logger()

    return log


def get_user_hostname() -> dict:
    """
    Retrieve user, hostname, IP, and OS configurations

    :return: ``sys_info``
    """

    sys_info = dict()

    sys_info['user'] = getuser()
    sys_info['hostname'] = gethostname()
    sys_info['ip'] = get('https://api.ipify.org').text

    os_name = uname()
    sys_info['os'] = f"{os_name[0]} {os_name[2]} {os_name[3]}"

    return sys_info


def get_log_file(log_handler) -> str:
    """
    Get log file

    :param log_handler: Logger object

    :return log_file: Full path of log file
    """
    log_file = ''
    if isinstance(log_handler, logging.FileHandler):
        log_file = log_handler.baseFilename
    return log_file


class LogCommons:
    """
    Common methods used when logging

    :param log: Logging object
    :param script_name: Name of script for log messages
    :param gi: Object containing git info
    :param code_name: Name of codebase/software (e.g., ReQUIAM, LD-Cool-P)
    :param version: Version of codebase/software. Default: Use ``redata``'s

    :ivar log: Logging object
    :ivar script_name: Name of script for log messages
    :ivar gi: Object containing git info
    :ivar code_name: Name of codebase/software (e.g., ReQUIAM, LD-Cool-P)
    :ivar version: Version of codebase/software.
    :ivar start_text: Text for script start
    :ivar asterisk: Parsing of start_text as asterisks
    :ivar sys_info: System info dict
    """

    def __init__(self, log: logging.Logger, script_name: str, gi: GitInfo,
                 code_name: str = '', version: str = __version__):
        self.log = log
        self.script_name = script_name
        self.gi = gi
        self.code_name = code_name
        self.version = version

        self.start_text: str = f"Started {script_name} script ... "
        self.asterisk: str = "*" * len(self.start_text)
        self.sys_info: dict = get_user_hostname()

    def script_start(self):
        """Log start of script"""
        self.log.info(self.asterisk)
        self.log.info(self.start_text)
        self.log.debug(f"{self.code_name} active branch: {self.gi.branch}")
        self.log.debug(f"{self.code_name} version: {self.version}"
                       f"({self.gi.short_commit})")
        self.log.debug(f"{self.code_name} commit hash: {self.gi.commit}")

    def script_sys_info(self):
        """Log system info"""
        self.log.debug(f"username : {self.sys_info['user']}")
        self.log.debug(f"hostname : {self.sys_info['hostname']}")
        self.log.debug(f"IP Addr  : {self.sys_info['ip']}")
        self.log.debug(f"Op. Sys. : {self.sys_info['os']}")

    def script_end(self):
        """Log end of script"""
        self.log.info(self.asterisk)
        self.log.info("Exit 0")

    def log_permission(self):
        """Change permission for file logs"""
        for handler in self.log.handlers:
            log_file = get_log_file(handler)
            if log_file:
                self.log.debug(f"Changing permissions for {log_file}")
                chmod(log_file, mode=0o666)


def pandas_write_buffer(df: pd.DataFrame, log_filename: str):
    """
    Write pandas content via to_markdown() to log_filename

    :param df: DataFrame to write to buffer
    :param log_filename: Full path for log file
    """

    buffer = io.StringIO()
    df.to_markdown(buffer)
    print(buffer.getvalue())  # This log to stdout

    with open(log_filename, mode='a') as f:
        print(buffer.getvalue(), file=f)
    buffer.close()


def log_settings(vargs: dict, config_dict: dict, protected_keys: list,
                 log: logging.Logger = log_stdout()) -> int:
    """
    Log parsed arguments settings for scripts

    :param vargs: Parsed arguments
    :param config_dict: Contains configuration settings. See commons.dict_load
    :param protected_keys: list of private arguments to print unset or set status
    :param log: LogClass

    :return: Number of errors with credentials
    """

    if log is None:
        log = log_stdout()

    cred_err = 0
    sections = config_dict.keys()
    for p in vargs.keys():
        value = ''
        for section in sections:
            if p in config_dict[section].keys():
                value = config_dict[section][p]

        if p in protected_keys:
            if value == '***override***' or value == '':
                log.info(f'   {p: >17} = (unset)')
                cred_err += 1
            else:
                log.info(f'   {p: >17} = (set)')
        else:
            log.info(f"   {p: >17} = {value}")

    return cred_err



from setuptools import setup, find_namespace_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fr:
    requirements = fr.read().splitlines()

setup(
    name='redata',
    version='0.4.2',
    packages=find_namespace_packages(),
    url='https://github.com/UAL-RE/redata-commons/',
    project_urls={
        'Source': 'https://github.com/UAL-RE/redata-commons/',
        'Tracker': 'https://github.com/UAL-RE/redata-commons/issues',
    },
    license='MIT License',
    author='Chun Ly',
    author_email='astro.chun@gmail.com',
    description='Commons code used by ReDATA software',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.7',
    install_requires=requirements,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ]
)
